//Program created by Ian Sunyecz 12/2020
//Farkle is a fun and challenging dice game with multiple variations and scoring options. The exact rules for the game can be found on
//https://www.dicegamedepot.com/farkle-rules/. This code works with my custom made Farkle PCB that includes an arduino pro mini 5V, a 4 digit 7 segment display,
//a 128x64 oled screen, 6x LED dice, and 6x capacitative touch sensors, among other components. For more information on the PCB design, components used, and
//playing the game, check out my instructables page: https://www.instructables.com/member/Sunyecz22/instructables/.

//Include required libraries to control MAX7219 chips, oled screen, 7 segment display
#include "LedControl.h"
#include "binary.h"
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels
// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
#define OLED_RESET     A3 // Reset pin # (or -1 if sharing Arduino reset pin)
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

//Define and name 7 segment and dice that are controlled by separate MAX7219 chips
//More info on controlling the 7 segment and dice LED's can be found from this
//awesome website: http://wayoda.github.io/LedControl/index.html
LedControl lc7seg = LedControl(13, 12, 11, 2); //7 segment control
LedControl lcDice = LedControl(10, 9, 8, 2); //Dice control

//Define Game Variables
//Hardware --> arduino pin values
const int buzzer = 2; //Piezo buzzer to Pin 2
const int RollB = 3; //Roll push button to Pin 3
const int Pass = 4; //Pass push button to Pin 4
const int Cap0 = 5; //Capacitative touch sensor 1 to pin 5
const int Cap1 = 6; //Capacitative touch sensor 2 to pin 6
const int Cap2 = 7; //Capacitative touch sensor 3 to pin 7
const int Cap3 = A0; //Capacitative touch sensor 4 to pin A0
const int Cap4 = A1; //Capacitative touch sensor 5 to pin A1
const int Cap5 = A2; //Capacitative touch sensor 6 to pin A2

//Bitmap declaration
extern uint8_t farkleBMB[];
extern uint8_t farkleBMW[];
extern uint8_t welcomeFarkleB[];
extern uint8_t welcomeFarkleW[];

//Variables used in calculating dice point value per turn
int scoreArray[6];
int scoreArrayOrig[6];
int updateArray[6];
int PrevScore = 0;
int oledPrevScore[6];
bool straightCheck = false;
bool sixCheck = false;
bool fiveCheck = false;
bool fourCheck = false;
bool threeCheck = false;
bool pairsCheck = false;
uint8_t fourIndex = 0;
uint8_t prevfourIndex = 0;
uint8_t threeIndex = 0;
uint8_t prevthreeIndex = 0;
uint8_t scoreCount = 0;
int pointValue = 0;

//Other variables used in game play
uint8_t NP;
uint8_t counter = 9;
uint8_t CapVal[6] = {Cap0, Cap1, Cap2, Cap3, Cap4, Cap5};
uint8_t CapSens[6];
uint8_t prevCapSens[6];
uint8_t rollVal[6];
uint8_t diceSelected = 0;
uint8_t currentPage;

//Game is won when player scores > 10,000 points
int winScore = 10000;
int winPlayer = 0;

//Boolean operators for various game functions
bool gameStart = false;
bool FarkleCheck = false;
bool win = false;
bool soundON = false; //Change to true if user wants auditory beep after each button press

//Current player variable (should change this variable to be something more meaningful than k)
uint8_t k = 0;
//Initialize 8 bit binary value for each die face
//Die Value      0           1         2           3           4           5        6
const int die[7] = {B00000000, B00000001, B00010010, B00010011, B01011010, B01011011, B01111110};
//Initialize 8 bit binary value for roll animation
const int roll[6] = {B01000000, B00100000, B00010000, B00001000, B00000100, B00000010};

void setup() {
  Serial.begin(9600);
  randomSeed(analogRead(A6)); //Generate random seed to use when rolling die for "random" outcome
  for (int i = 0; i <= 5; i++) { //Initialize capacitate touch sensors
    pinMode("Cap" + i, INPUT);
  }
  pinMode(RollB, INPUT); //Roll push button
  pinMode(Pass, INPUT); //Pass push button
  pinMode(buzzer, OUTPUT); //Piezo buzzer
  //Max7219 chips initialization
  lc7seg.shutdown(0, false); //MAX7219 chips into normal operation
  lcDice.shutdown(0, false);
  lc7seg.setIntensity(0, 8); //Brightness
  lcDice.setIntensity(0, 4);
  lc7seg.clearDisplay(0); //Clear the display
  lcDice.clearDisplay(0);
  delay(100);
  //Set up 128x64 OLED
  if (!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)) { // Address 0x3D for 128x64
    Serial.println(F("SSD1306 allocation failed"));
    for (;;); // Don't proceed, loop forever
  }
  display.clearDisplay();
  display.display();
  welcomeScreen(); //Run welcome screen on startup
  numplayScreen(); //Run "number of players" screen after welcom
  currentPage = '0'; //Set current page to 0
}

void loop() {
  //Page 0 is the page where user can select the number of players in the game. Currently, game allows 2-6 players.
  //There is a setting for 1 player but this is not coded yet. My idea is to have some type of "computer" play the
  //user if 1 player is selected. That will be in version 2.0 :p
  if (currentPage == '0') {
    numberSelect(); //Select total number of players in game (NP is the variable to store number of players)
    int RollBVal = digitalRead(RollB);
    //Player presses roll push button to begin game after # of players are selected
    if (RollBVal == HIGH) {
      beep();
      currentPage = '1';
      scrollDigits();
      playerScreen();
      lcDice.clearDisplay(0);
      playerScore(0);
      playerHighlight(0);
      printNumber(0);
    }
  }

  //Page 1 is the page that begins each players round. Player must hit 'roll' to roll first set of dice
  if (currentPage == '1') {
    //Checks to see if there is a winner at beginning of new round. Allows one final round to be played should that be the scenario
    if (k == 0) {
      for (int x = 0; x < NP; x++) {
        //Checks for score >10,000 but also checks for the highest score should multiple players have > 10,000
        if (oledPrevScore[x] > winScore) {
          winScore = oledPrevScore[x];
          winPlayer = x;
          win = true; //Change state of win to true. Indicates a player has won
        }
      }
      if (win == true) {
        winner(winPlayer); //Winner animation
        win = false;
        winScore = 10000; //Reset win score to 10,000
      }
    }
    //Player presses roll push button to roll dice and begin their round
    int RollBVal = digitalRead(RollB);
    if (RollBVal == HIGH) {
      beep();
      diceAnimation();
      rollDice();
      currentPage = '2';
      pointValue = 0;
      PrevScore = 0;
    }
  }

  //Page 2 is the game play page. Player can select/deselect dice during this round. Points are tallied and are updated to the
  //7 segment display in real time. Player then has the option to "Roll" which rolls the remaining die/rolls 6 new die if all die were
  //selected. Or player has the option to "Pass" which logs the players score onto the oled and next player in line begins his/her turn.
  //Additionally, prior to either "rolling" or "passing", the program checks to see if a Farkle has occured. If so, no points are tallied for that round
  if (currentPage == '2') {
    int RollBVal = digitalRead(RollB);
    int PassVal = digitalRead(Pass);
    chooseDice(); //Main function that allows user to select/deselect dice on the game board
    checkScore(); //Main function that calculates current points in real time and updates points onto 7 segment display
    printNumber(pointValue + PrevScore); //Display current score onto 7 segment display
    delay(50);
    if (PassVal == HIGH || FarkleCheck == true) { //If player selects "Pass" push button: check for farkle, update score on oled screen, send back to Page 1 for next player
      if (FarkleCheck == false) {
        farkle(k); //Check for a Farkle first
        if (FarkleCheck == false) {
          tone(buzzer, 500);
          delay(50);
          tone (buzzer, 700);
          delay(50);
          tone(buzzer, 900);
          delay(50);
          noTone(buzzer);
        }
      }
      playerScreen();
      playerScore(k); //Update score on oled screen
      if (k + 1 == NP) {
        playerHighlight(0);
        k = 0;
      }
      else {
        playerHighlight(k + 1);
        k++;
      }
      printNumber(0);
      lcDice.clearDisplay(0);
      //Reset scoring arrays for next player/round
      for (int j = 0; j < 6; j++) {
        scoreArrayOrig[j] = 0;
        scoreArray[j] = 0;
        prevCapSens[j] = 0;
      }
      pointValue = 0;
      PrevScore = 0;
      FarkleCheck = false;
      currentPage = '1';
    }
    if (RollBVal == HIGH) { //If player selects "Roll" push button: Check for farkle, roll remaining die/or 6 new die should all die be selected
      beep();
      farkle(k);
      if (FarkleCheck == true) {
        lcDice.clearDisplay(0);
      }
      else {
        lastPlay();
      }
      oledPrevScore[k] = pointValue + oledPrevScore[k];
    }
    pointValue = 0;
  }
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
////---------------------------------------------------------------------------Start of Game Functions----------------------------------------------------------------------////
////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////

//chooseDice() function allows the user to 'select' and 'deselect' a die when touching the capacitative
//sensor on the game board. If a user 'selects' a die, it will clear from the board. If the user
//'selects' that same die again, it will reappear on the game board with the same die value as before
void chooseDice () {
  for (int i = 0; i < 6; i++) {
    CapSens[i] = digitalRead(CapVal[i]);
  }
  for (int j = 0; j < 6; j++) {
    if (CapSens[j] == HIGH) {
      beep();
      if (CapSens[j] != prevCapSens[j]) { //If sensor is touched, turn die LEDs off, and store whatever was displayed on the die in "scoreArrayOrig"
        lcDice.setRow(0, j, die[0]);
        scoreArrayOrig[j] = rollVal[j];
        delay(500);
      }
      if (CapSens[j] == prevCapSens[j]) { //If sensor was off and sensor was touched again, turn die LEDs on, and make that value in "scoreArrayOrig" = 0
        lcDice.setRow(0, j, die[rollVal[j]]);
        scoreArrayOrig[j] = 0;
        delay(500);
      }
      if (CapSens[j] == HIGH && prevCapSens[j] == 1) {
        prevCapSens[j] = 0;
        break;
      }
      if (CapSens[j] == HIGH && prevCapSens[j] == 0) {
        prevCapSens[j] = CapSens[j];
        break;
      }
    }
  }
}
////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////

//checkScore() is the brains behind this game. It is the function that calculates the current score based on the dice that the user has selected. This website,
//https://www.dicegamedepot.com/farkle-rules/ contains the rules for Farkle and is how I based this function. Due to the massive number of scoring variations and to
//make things a bit simpler, I arrange all 6 die values (if not selected by player, value = 0) into ascending order. I then check for individual dice combinations realizing
//that multiple scoring combinations cannot exist at the same time. For example, a player cannot obtain points for 6 of a kind and 4 of a kind in the same roll. If the dice
//combination selected does not fit a certian scoring criteria, no points are awarded, and the dice combination is screened to the next potential point combination.
void checkScore() {
  //First arrange selected dice combination into ascending order
  for (int j = 0; j < 6; j++) {
    scoreArray[j] = scoreArrayOrig[j];
  }
  int scoreArray_length = sizeof(scoreArray) / sizeof(scoreArray[0]);
  qsort(scoreArray, scoreArray_length, sizeof(scoreArray[0]), sort_asc);
  for (int i = 0; i < 6; i++) {
    updateArray[i] = scoreArray[i];
  }

  //----------------------------------------------
  //Check for straight 1-6
  for (int i = 0; i < 6; i++) {
    if (scoreArray[i] == i + 1) {
      scoreCount++;
    }
  }
  if (scoreCount == 6) {
    straightCheck = true;
    pointValue = 1500;
  }
  //----------------------------------------------
  //Check for 6 of a Kind
  scoreCount = 0;
  if (straightCheck == false) {
    for (int i = 0; i < 5; i++) {
      if (scoreArray[i] == scoreArray[i + 1] && scoreArray[1] != 0) {
        scoreCount++;
      }
    }
    if (scoreCount == 5) {
      pointValue = 3000;
      sixCheck = true;
    }
  }
  //----------------------------------------------
  //Check for 5 of a kind
  //Implmenting "updateArray" here; Will replace the scored integers with zeros which should aid in solving for additional points after 5,4, or 3 of kind is solved;
  //I.e. scoreArray = [4, 4, 4, 4, 5, 5] Points will be given for the 4 of a kind but the "updateArray" will display [0,0,0,0,5,5]; Then if 5 or 4 of a kind is awarded,
  //the new array can be checked for additional points in the form of "1" or "5"
  scoreCount = 0;
  if (scoreArray[0] < scoreArray[1]) { //Must check first integer; If larger than integer of 5 of a kind, will be sorted to scoreArray[5]
    for (int i = 1; i < 5; i++) {
      if (scoreArray[i] == scoreArray[i + 1] && scoreArray[1] != 0 && scoreArray[1] != -1 && straightCheck == false && sixCheck == false) {
        pointValue = 0;
        scoreCount++;
        updateArray[i] = 0;
        if (scoreCount == 4) {
          updateArray[i + 1] = 0;
        }
      }
    }
    if (scoreCount == 4) {
      pointValue = 2000;
      fiveCheck = true;
    }
  }
  else { //Must check first integer; If larger than integer of 5 of a kind, will be sorted to scoreArray[5]
    for (int i = 0; i < 5; i++) {
      if (i == 4) {
        break;
      }
      if (scoreArray[i] == scoreArray[i + 1] && scoreArray[0] != 0 && scoreArray[0] != -1 && sixCheck == false) {
        scoreCount++;
        updateArray[i] = 0;
        if (scoreCount == 4) {
          updateArray[i + 1] = 0;
        }
      }
    }
    if (scoreCount == 4) {
      pointValue = 2000;
      fiveCheck = true;
    }
  }

  if (fiveCheck == true) {
    for (int i = 0; i < 6; i++) {
      updateArray[i] = scoreArray[i];
    }
  }
  //----------------------------------------------
  //Check for 4 of a kind
  scoreCount = 0;
  if (straightCheck == false && sixCheck == false && fiveCheck == false) {
    pointValue = 0;
    for (int i = 0; i < 5; i++) {
      if (scoreArray[i] != 0 && scoreArray[i] != -1) {
        if (scoreArray[i] == scoreArray[i + 1]) {
          scoreCount++;
          updateArray[i] = 0;
          fourIndex = scoreArray[i];
          if (scoreCount == 3) {
            updateArray[i + 1] = 0;
            break;
          }
        }
        //This checks to see if there are 2 of a kind before 4 of a kind e.g. (1, 1, 4, 4, 4, 4)
        else {
          if (scoreCount <= 2) {
            scoreCount = 0;
          }
          if (i > 0) {
            updateArray[i - 1] = scoreArray[i - 1];
          }
        }
      }
    }
    if (scoreCount == 3) {
      pointValue = 1000;
      fourCheck = true;
    }

    if (fourCheck == false) {
      for (int i = 0; i < 6; i++) {
        updateArray[i] = scoreArray[i];
      }
    }
  }
  //---------------------------------------------
  //Check for 3 of a kind
  scoreCount = 0;
  if (straightCheck == false && sixCheck == false && fiveCheck == false && fourCheck == false) {
    pointValue = 0;
    for (int i = 0; i < 5; i++) {
      if (scoreArray[i] != 0 && scoreArray[i] != -1) {
        if (scoreArray[i] == scoreArray[i + 1]) {
          scoreCount++;
          threeIndex = scoreArray[i];
          if (scoreCount == 2) {
            updateArray[i] = 0;
            updateArray[i + 1] = 0;
            updateArray[i - 1] = 0;
          }
        }
        //First 2 numbers must not be 3 of a kind e.g. (1, 1, 4, 4, 4, 5)
        if (scoreArray[i] != scoreArray[i + 1] && scoreCount <= 1) {
          scoreCount = 0;
        }
        //This now checks to see if there is another 3 of a kind e.g. (1, 1, 1, 4, 4, 4)
        //by indexing the first 3
        if (i == 2 && scoreCount == 2) {
          prevthreeIndex = scoreArray[2];
          scoreCount = 0;
        }
      }
    }
    //One triplet in the sequence
    if (scoreCount == 2 && prevthreeIndex == 0) {
      if (threeIndex != 1) {
        pointValue = threeIndex * 100;
      }
      else {
        pointValue = 300;
      }
      threeCheck = true;
    }
    //Two triplets in sequence which = 2500;
    else if (scoreCount == 2 && prevthreeIndex != 0) {
      pointValue = 2500;
      threeCheck = true;
    }
    //1 triplet in sequence but as first 3 number in sequence
    else if (scoreCount == 0 && prevthreeIndex != 0) {
      if (prevthreeIndex != 1) {
        pointValue = prevthreeIndex * 100;
      }
      else {
        pointValue = 300;
      }
      threeCheck = true;
    }
    else if (scoreCount == 1 && prevthreeIndex != 0) {
      if (prevthreeIndex != 1) {
        pointValue = prevthreeIndex * 100;
      }
      else {
        pointValue = 300;
      }
      threeCheck = true;
    }
  }
  //---------------------------------------------
  //Check for 3 pairs e.g. {1, 1, 2, 2, 3, 3,}
  scoreCount = 0;
  if (straightCheck == false && sixCheck == false && fiveCheck == false && fourCheck == false && threeCheck == false) {
    pointValue = 0;
    for (int i = 0; i < 6; i = i + 2) {
      if (scoreArray[i] != 0 && scoreArray[i] != -1) {
        if (scoreArray[i] == scoreArray[i + 1]) {
          scoreCount++;
          updateArray[i] = 0;
          updateArray[i + 1] = 0;
        }
      }
    }

    if (scoreCount == 3) {
      pointValue = 1500;
      pairsCheck = true;
    }
  }
  //---------------------------------------------
  //Check for either 1 or 5's
  if (straightCheck == false && sixCheck == false && fiveCheck == false && fourCheck == false && threeCheck == false && pairsCheck == false) {
    for (int i = 0; i < 6; i++) {
      if (scoreArray[i] == 1) {
        pointValue = pointValue + 100;
      }
      if (scoreArray[i] == 5) {
        pointValue = pointValue + 50;
      }
    }
  }
  if (fiveCheck == 1 || fourCheck == 1 || threeCheck == 1 || pairsCheck == 1) {
    for (int i = 0; i < 6; i++) {
      if (updateArray[i] == 1) {
        pointValue = pointValue + 100;
      }
      if (updateArray[i] == 5) {
        pointValue = pointValue + 50;
      }
    }
  }

  //Reset Values
  scoreCount = 0;
  straightCheck = false;
  sixCheck = false;
  fiveCheck = false;
  fourCheck = false;
  threeCheck = false;
  pairsCheck = false;
  fourIndex = 0;
  prevfourIndex = 0;
  threeIndex = 0;
  prevthreeIndex = 0;
  for (int j = 0; j < 6; j++) {
    scoreArray[j] = 0;
  }
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//LastPlay() is run if user selects "Roll" push button to continue round. This function saves the amount of points tallied from previous roll, and rolls new dice for the
//remaining die left on the board. Should all 6 dice be selected and user wants to continue round, this function will save the points and roll 6 new die.
void lastPlay() {
  PrevScore = pointValue + PrevScore; // Updates score from previous round
  lcDice.clearDisplay(0);

  diceAnimation();

  lcDice.clearDisplay(0);
  for (int j = 0; j < 6; j++) {
    if (scoreArrayOrig[j] == 0) { //If dice were not selected, roll a new value for those die and display it
      rollVal[j] = random(1, 7);
      lcDice.setRow(0, j, die[rollVal[j]]);
    }
    else { //If dice were selected, store it as arbitrary -1 in the "scoreArrayOrig" variable for scoring purposes so that it does not get counted into the subsequent round(s)
      scoreArrayOrig[j] = -1;
    }
  }
  //Counts the die on the board to see how many have already been selected
  for (int i = 0; i < 6; i++) {
    if (scoreArrayOrig[i] == -1) {
      diceSelected++;
    }
  }
  //If all 6 dice were selected, clear board and scoring variables, roll all dice again, keep points
  if (diceSelected == 6) {
    for (int m = 0; m < 6; m++) {
      scoreArrayOrig[m] = 0;
      scoreArray[m] = 0;
      prevCapSens[m] = 0;
    }
    diceSelected = 0;
    diceAnimation();
    rollDice();
  }
  //If all dice were not selected reset this tally variable so it is not cumulative and proceed with round
  else {
    diceSelected = 0;
  }
  delay(50);
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//printNumber(x) function prints numbers to the 7 segment display. Modified from http://wayoda.github.io/LedControl/index.html
void printNumber(int v) {
  int ones;
  int tens;
  int hundreds;
  int thousands;
  boolean negative = false;
  if (v < -999 || v > 9999)
    return;
  if (v < 0) {
    negative = true;
    v = v * -1;
  }
  ones = v % 10;
  v = v / 10;
  tens = v % 10;
  v = v / 10;
  hundreds = v % 10;
  //
  v = v / 10;
  thousands = v;
  //
  if (negative) {
    //print character '-' in the leftmost column
    lc7seg.setChar(0, 3, '-', false);
  }
  else {
    //print a blank in the sign column
    lc7seg.setChar(0, 3, ' ', false);
  }
  //Now print the number digit by digit
  lc7seg.setDigit(0, 3, (byte)thousands, false);
  lc7seg.setDigit(0, 2, (byte)hundreds, false);
  lc7seg.setDigit(0, 1, (byte)tens, false);
  lc7seg.setDigit(0, 0, (byte)ones, false);
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//rollDice() function creates a random number for each die and displays that number onto the die LED's
void rollDice () {
  for (int l = 0; l < 6; l++) {
    rollVal[l] = random(1, 7);
    lcDice.setRow(0, l, die[rollVal[l]]);
  }
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//diceAnimation() creates an animation when rolling the dice. Only run animation on die remaining on the board
void diceAnimation() {

  for (int i = 0; i < 6; i++) {
    for (int j = 0; j < 6; j++) {
      if (prevCapSens[j] == 1) {
        lcDice.setRow(0, j, B00000000);
      }
      else {
        lcDice.setRow(0, j, roll[i]);
      }
    }
    delay(100);
  }
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//scrollDigits() function scrolls the digits on the die face. This animation is run once at start of game after the player hits "Roll" to begin the game.
void scrollDigits() {
  for (int i = 1; i < 7; i++) {
    lcDice.setRow(0, 0, die[i]);
    lcDice.setRow(0, 1, die[i]);
    lcDice.setRow(0, 2, die[i]);
    lcDice.setRow(0, 3, die[i]);
    lcDice.setRow(0, 4, die[i]);
    lcDice.setRow(0, 5, die[i]);
    delay(200 + (15 * i));
  }
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//welcomeScreen() function is the first screen to display on the old welcoming users to the game. "Welcome to Farkle"
void welcomeScreen() {
  display.clearDisplay();
  display.setTextSize(2);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(29, 24);
  display.print(F("WELCOME"));
  display.display();
  delay(50);
  tone(buzzer, 400);
  delay(50);
  tone (buzzer, 450);
  delay(50);
  tone(buzzer, 500);

  display.invertDisplay(true);
  delay(200);
  display.invertDisplay(false);
  delay(200);
  display.invertDisplay(true);
  delay(200);
  display.invertDisplay(false);

  tone(buzzer, 600);
  delay(50);
  tone (buzzer, 650);
  delay(50);
  tone(buzzer, 700);
  display.clearDisplay();
  display.setCursor(54, 24);
  display.print(F("TO"));
  display.display();
  delay(50);
  display.invertDisplay(true);
  delay(200);
  display.invertDisplay(false);
  delay(200);
  display.invertDisplay(true);
  delay(200);
  display.invertDisplay(false);

  tone(buzzer, 800);
  delay(50);
  tone (buzzer, 850);
  delay(50);
  tone(buzzer, 900);
  for (int i = 0; i < 5; i++) {
    display.clearDisplay();
    display.drawBitmap(0, 0, welcomeFarkleW, 128, 64, 1);
    display.display();
    delay(100);
    display.clearDisplay();
    display.drawBitmap(0, 0, welcomeFarkleB, 128, 64, 1);
    display.display();
    delay(100);
  }
  noTone(buzzer);
  delay(2000);
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//numplayScreen() function prompts the user to select the number of players by writing to the old screen and displaying 1-6 on the die faces
void numplayScreen() {
  display.clearDisplay();
  display.setTextColor(SSD1306_WHITE);
  display.setTextSize(2);
  display.setCursor(24, 0);
  display.print(F("HOW MANY"));
  display.setCursor(24, 17);
  display.print(F("PLAYERS?"));
  display.drawLine(0, 35, 128, 35, SSD1306_WHITE);
  display.setTextSize(1);
  display.setCursor(10, 37);
  display.print(F("Touch corresponding"));
  display.setCursor(13, 46);
  display.print(F("dice pad to select"));
  display.setCursor(13, 55);
  display.print(F("number of players!"));
  display.display();
  lcDice.clearDisplay(0);
  lcDice.setRow(0, 0, die[1]); // Dice 1
  lcDice.setRow(0, 1, die[2]); // Dice 2
  lcDice.setRow(0, 2, die[3]); // Dice 3
  lcDice.setRow(0, 3, die[4]); // Dice 4
  lcDice.setRow(0, 4, die[5]); // Dice 5
  lcDice.setRow(0, 5, die[6]); // Dice 6
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//numberSelect() function runs on startup when the user is selecting the number of players. It works with
//the diceslectScreen() function but this specifically changes the number of players on the oled screen as the
//player selects 1-6
void numberSelect() {
  for (int i = 0; i < 6; i++) {
    CapSens[i] = digitalRead(CapVal[i]);
  }
  for (int j = 0; j < 6; j++) {
    if (CapSens[j] == HIGH) {
      beep();
      NP = j + 1;
      diceselectScreen();
      display.setTextSize(2);
      display.setCursor(59, 0);
      display.print(NP);
      display.display();
      delay(500);
    }
    else {
    }
  }
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//diceselectScreen() function will update the oled screen with the number of players selected when the player chooses the corresponding capacitative touch pad
void diceselectScreen() {
  display.clearDisplay();
  display.setTextColor(SSD1306_WHITE);
  display.setTextSize(2);
  display.setCursor(29, 15);
  display.print(F("Players"));
  display.drawLine(0, 34, 128, 34, SSD1306_WHITE);
  display.setTextSize(1);
  display.setCursor(17, 36);
  display.print(F("Please hit 'Roll'"));
  display.setCursor(15, 45);
  display.print(F("when player number"));
  display.setCursor(32, 54);
  display.print(F("is selected"));
  display.display();
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//playerScreen() function creates the player score screen that will be displayed during game play.
//The number of players on the screen is dictated by what the user selects in the start up prompt
void playerScreen() {
  display.clearDisplay();
  display.setTextSize(1);// Normal 1:1 pixel scale
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(46, 0);
  display.print(F("FARKLE!"));
  display.drawLine(0, 9, 128, 9, SSD1306_WHITE);
  display.setCursor(0, 11);
  for (int i = 1; i <= NP; i++) {
    display.print(F("Player "));
    display.print(i);
    display.print(F(":"));
    display.setCursor(0, 11 + counter);
    counter = counter + 9;
  }
  display.display();
  counter = 9;
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//playerHighlight(x) function highlights the current player on the oled screen
void playerHighlight(int x) {
  int Cursor[6] = {11, 20, 29, 38, 47, 56};
  display.setTextColor(SSD1306_BLACK, SSD1306_WHITE); // Draw 'inverse' text
  display.setCursor(0, Cursor[x]);
  display.print(F("Player "));
  display.print(x + 1);
  display.print(F(":"));
  display.display();
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//playerScore(x) function updates the current players score on the oled screen. This function accounts for a Farkle and will update and/or keep the score on the oled accordingly
void playerScore(int x) {
  int Cursor[6] = {11, 20, 29, 38, 47, 56};
  if (gameStart == false) { //Run once at the beginning to give every player 0 points initially
    for (int j = 0; j < NP; j++) {
      display.setCursor(70, Cursor[j]);
      display.print(0);
    }
    display.display();
    gameStart = true;
  }
  else {
    if (FarkleCheck == false) { //No Farkle: Update current players points and keep the score of the remaining players the same
      for (int j = 0; j < NP; j++) {
        if (j == x) {
          display.setCursor(70, Cursor[j]);
          display.print(pointValue + oledPrevScore[j]);
          display.display();
          oledPrevScore[j] = pointValue + oledPrevScore[j];
        }
        else {
          display.setCursor(70, Cursor[j]);
          display.print(oledPrevScore[j]);
          display.display();
        }
      }
    }
    else { //Farkle: No points awarded to current player. Display score from previous round and keep the score of the remaining players the same
      for (int j = 0; j < NP; j++) {
        if (j == x) {
          oledPrevScore[j] = oledPrevScore[j] - PrevScore;
          display.setCursor(70, Cursor[j]);
          display.print(oledPrevScore[j]);
          display.display();
        }
        else {
          display.setCursor(70, Cursor[j]);
          display.print(oledPrevScore[j]);
          display.display();
        }
      }
    }
  }
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//farkle(x) function checks for a farkle if no points were awarded for any dice combination. Play loser buzzer and display "Farkle" bitmap on oled
void farkle(int x) {
  if (pointValue == 0) { //Farkle and don't add up any points
    FarkleCheck = true;
    tone(buzzer, 1000);
    delay(300);
    tone(buzzer, 400);
    for (int i = 0; i < 5; i++) {
      display.clearDisplay();
      display.drawBitmap(32, 0, farkleBMW, 64, 64, 1);
      display.display();
      delay(100);
      display.clearDisplay();
      display.drawBitmap(32, 0, farkleBMB, 64, 64, 1);
      display.display();
      delay(100);
    }
    noTone(buzzer);
    delay(1000);
  }
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//winner(x) function updates oled with winning player, plays winning sound from piezo, and creates an animation on the oled. Returns to beginning of game once completed
void winner(int x) {
  delay(500);
  display.clearDisplay();
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(49, 0);
  display.print(F("PLAYER"));
  display.setCursor(62, 12);
  display.print(x + 1);
  display.setTextSize(2);
  display.setCursor(24, 30);
  display.print(F("WINNER!!"));
  display.display();
  display.setTextSize(1);
  //Play fun winner tone
  tone(buzzer, 800);
  delay(125);
  noTone(buzzer);
  delay(25);
  tone(buzzer, 900);
  delay(125);
  noTone(buzzer);
  delay(25);
  tone(buzzer, 1000);
  delay(125);
  noTone(buzzer);
  delay(25);
  tone(buzzer, 1200);
  delay(225);
  noTone(buzzer);
  delay(25);
  tone(buzzer, 900);
  delay(125);
  noTone(buzzer);
  delay(25);
  tone(buzzer, 1000);
  delay(125);
  noTone(buzzer);
  delay(25);
  tone(buzzer, 1200);
  delay(500);
  noTone(buzzer);
  delay(500);
  tone(buzzer, 1195);
  delay(125);
  noTone(buzzer);
  delay(25);
  tone(buzzer, 1190);
  delay(1000);
  noTone(buzzer);
  //Fun winner animation with faux fireworks on the screen
  for (int z = 0; z < 8; z++) {
    display.clearDisplay();
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(49, 0);
    display.print(F("PLAYER"));
    display.setCursor(62, 12);
    display.print(x + 1);
    display.setTextSize(2);
    display.setCursor(24, 30);
    display.print(F("WINNER!!"));
    display.display();
    display.setTextSize(1);
    int randomFW_Y = random(40, 60);
    int randomFW_X1 = random(5, 40);
    int randomFW_X2 = random(50, 90);
    int randomFW_X3 = random(100, 120);
    for (int y = randomFW_Y; y > 10; y--) {
      display.drawPixel(randomFW_X1, y, SSD1306_WHITE);
      display.drawPixel(randomFW_X2, y + 10, SSD1306_WHITE);
      display.drawPixel(randomFW_X3, y - 5, SSD1306_WHITE);
      display.display();
      delay(5);
      display.drawPixel(randomFW_X1, y, SSD1306_BLACK);
      display.drawPixel(randomFW_X2, y + 10, SSD1306_BLACK);
      display.drawPixel(randomFW_X3, y - 5, SSD1306_BLACK);
      delay(5);
    }
    for (int x = 1; x < 10; x++) {
      display.drawCircle(randomFW_X1, 11, x, SSD1306_WHITE);
      display.drawCircle(randomFW_X2, 21, x, SSD1306_WHITE);
      display.drawCircle(randomFW_X3, 6, x, SSD1306_WHITE);
      display.display();
      delay(5);
      display.drawCircle(randomFW_X1, 11, x, SSD1306_BLACK);
      display.drawCircle(randomFW_X2, 21, x, SSD1306_BLACK);
      display.drawCircle(randomFW_X3, 6, x, SSD1306_BLACK);
      display.display();
    }
  }
  delay(1000);
  for (int x = 0; x < NP; x++) {
    oledPrevScore[x] = 0;
  }
  currentPage = '0';
  k = 0;
  numplayScreen();
}

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//beep() adds a short beep when buttons on the PCB are pressed.
void beep() {
  if (soundON == true) {
    tone(buzzer, 500);
    delay(100);
    noTone(buzzer);
  }
}
////------------------------------------------------------------------------------------------------------------------------------------------------------------------------////
//This is needed in order to arrange dice combination arrays in ascending order
int sort_asc(const void *cmp1, const void *cmp2)
{
  // Need to cast the void * to int *
  int a = *((int *)cmp1);
  int b = *((int *)cmp2);
  // The comparison
  return a < b ? -1 : (a > b ? 1 : 0);
  // A simpler, probably faster way:
  //return b - a;
}
